# Data preprocess
import scanpy
import numpy as np
import pandas as pd


def qcFilter(adata):
    '''
    Remove non-expressed cells and genes with quality control.
    :param expr: (pd.DataFrame): single cell anndata.
    :return: (pd.DataFrame): Cell-by-gene expression matrix after quality control.
    '''
    scanpy.pp.filter_cells(adata, min_counts=1)
    scanpy.pp.filter_genes(adata, min_counts=1)
    return adata.to_df()

def findHVG(expr, num_HVG):
    '''
    Find highly variable genes (HVGs).
    :param expr: (pd.DataFrame): Cell-by-gene expression matrix.
    :param num_HVG: (int) The number of highly variable genes.
    :return: (pd.DataFrame): Cell-by-HVG expression matrix.
    '''
    adata = scanpy.AnnData(expr.copy())
    scanpy.pp.log1p(adata)
    hvg_idx = scanpy.pp.highly_variable_genes(adata, n_top_genes=num_HVG, inplace=False)
    if len(np.where(hvg_idx.highly_variable.values)[0]) > num_HVG:
        idx_by_mean = np.argsort(hvg_idx[hvg_idx.highly_variable == True].means.values)[::-1]
        hvg_idx = idx_by_mean[:num_HVG]
    else:
        hvg_idx = np.where(hvg_idx.highly_variable.values)[0]
    return expr.iloc[:, hvg_idx]

def min_max_normalization(data):
    """
    0-1 normalization(gene normalization)
    input data format: row is cell and col is gene
    """
    data = (data - data.min(axis=0))/(data.max(axis=0) - data.min(axis=0))
    return data.T



# Load the model to extract features to calculate correlations between genes
import torch
import torch as nn
import pandas as pd
import numpy as np
from torch.utils.data import Dataset
import sys
sys.path.append('../GeneCluster')
from utils import load_model

class ExpressionDataset(Dataset):
    '''
    custom Dataset class for classifier training.
    __len__: returns the number of samples in our dataset.
    __getitem__:  loads and returns a sample from the dataset at the given index idx.
    '''
    def __init__(self, data_path):
        '''
        label_file: label
        '''
        self.datas = pd.read_csv(data_path, sep='\t', index_col=0)    ##load datas
        self.datas = np.expand_dims(np.array(self.datas), axis=1)
        self.labels = np.zeros((len(self.datas)), dtype='float32')
    
    def __len__(self):
        return len(self.datas)

    def __getitem__(self, idx):
        data = self.datas[idx]
        label = self.labels[idx]
        data = torch.FloatTensor(data)
        label = torch.LongTensor(np.array(label))
        return data, label


def compute_features(dataloader, model, N):  
    model.eval()
    
    for batch, (input_tensor, _) in enumerate(dataloader): 
        with torch.no_grad():
            aux = model(input_tensor).data.cpu().numpy() 
        
        if batch == 0:
            features = np.zeros((N, aux.shape[1]), dtype='float32')  

        aux = aux.astype('float32')  
        if batch < len(dataloader) - 1:  
            features[batch * 128: (batch + 1) * 128] = aux
        else:
            # special treatment for final batch 
            features[batch * 128:] = aux

    return features


if __name__ == "__main__":
    expr_data_path = 'normal_data.csv'
    expr_data = pd.read_csv(expr_data_path,sep = '\t',index_col=0)
    dataset = ExpressionDataset(expr_data_path)
    dataloader = torch.utils.data.DataLoader(dataset,
                                            batch_size=128,
                                            num_workers=8,
                                            pin_memory=True)
    
    model = load_model('../train_res/checkpoint.pth.tar')
    model.top_layer = None
    model.classifier = nn.Sequential(*list(model.classifier.children())[:-1]) 
    features = compute_features(dataloader, model, len(dataset))
    features = pd.DataFrame(features,index=expr_data.index)
    features.to_csv('./features.csv')



import math
from typing import Tuple

import torch
from torch import nn, Tensor
import torch.nn.functional as F
from torch.nn import TransformerEncoder, TransformerEncoderLayer
from torch.utils.data import dataset

config = [(256, 3, 1, 1), (512, 3, 1, 1), (1024, 3, 1, 1),(1024, 3, 1, 1), (512, 3, 1, 1), (256, 3, 1, 1), "SPP"]
out_put_size = [1,3,6,8,10,16]

class SPPNet(nn.Module):
    def __init__(self, features, num_classes):
        super(SPPNet, self).__init__()
        self.features = features
        self.classifier = nn.Sequential(nn.Dropout(0.2),
                            nn.Linear(256 * sum(out_put_size), 2048), #加了一层空间金字塔池化
                            nn.ReLU(inplace=False),
                            nn.Dropout(0.2),
                            nn.Linear(2048, 2048),
                            nn.ReLU(inplace=False))
                            
        self.top_layer = nn.Linear(2048, num_classes)
        self._initialize_weights()
        
    def forward(self, x):
        x = self.features(x)
        x = self.classifier(x)
        if self.top_layer:
            x = self.top_layer(x)
        return x

    def _initialize_weights(self):
        for y, m in enumerate(self.modules()):
            if isinstance(m, nn.Conv1d):            
                n = m.kernel_size[0] * m.kernel_size[0] * m.out_channels
                for i in range(m.out_channels):
                    m.weight.data[i].normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm1d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                m.weight.data.normal_(0, 0.01)
                m.bias.data.zero_()

          

def make_layers_features(config, input_dim, bn):
    layers = []
    in_channels = input_dim
    for i in config:
        if i == 'MP':
            layers += [nn.MaxPool1d(kernel_size=3, stride=2)]
        elif i == 'SPP':
            layers += [SPP_1d(out_put_size)]
        else:
            conv1d = nn.Conv1d(in_channels, i[0], kernel_size=i[1],
                               stride=i[2], padding= i[3])
            if bn:
                layers += [conv1d, nn.BatchNorm1d(i[0]), nn.ReLU(inplace=False)]  
            else:
                layers += [conv1d, nn.ReLU(inplace=False)] 
            in_channels = i[0]
    return nn.Sequential(*layers)



class SPP_1d(nn.Module):
    '''
    1D Spatial pyramid pool layer
    '''
    def __init__(self, out_pool_size): 
        '''
        out_pool_size:The output size of each layer of the pool pyramid
        '''
        super(SPP_1d, self).__init__()
        self.out_pool_size = out_pool_size
        
    def forward(self, x):
        N, C, L = x.size() #N:Sample size， C:Number of channels， L:Input length
        for i in range(len(self.out_pool_size)):
            wid = int(math.ceil(L / self.out_pool_size[i]))   
            pad = wid * self.out_pool_size[i] - L    
            maxpool = nn.MaxPool1d(wid, stride=wid)
            x_pad = F.pad(x, (0, pad), "constant", 0)
            output = maxpool(x_pad)
            if (i == 0):
                spp = output.view(N, -1)
            else:
                spp = torch.cat((spp, output.view(N, -1)), 1)
        return spp
            

             
def sppCNN(bn=True, out=100):
    dim = 1
    model = SPPNet(make_layers_features(config, dim, bn=bn), out)
    return model


# model = sppCNN()
# input = torch.randn(20, 1, 758)
# a = model.features(input)
# print(a.size())    #[20,1536]
# print(a.type())
# print(a.type())   #[20,100]



import os
import pickle

import pandas as pd
import numpy as np

import torch
import models
from torch.utils.data import Dataset
from torch.utils.data.sampler import Sampler


def load_model(path):
    '''
    load model from checkpoint path.

    Parameters:
        path: checkpoint file path.
    Returns:
        model (torch.nn.Module): Loaded model or None if file not found.    
    '''
    if os.path.isfile(path):
        print("=> loading checkpoint '{}'".format(path))
        checkpoint = torch.load(path)

        # size of the top layer
        N = checkpoint['state_dict']['top_layer.bias'].size()

        # build skeleton of the model
        model = models.__dict__[checkpoint['arch']](out=int(N[0]))

        # deal with a dataparallel table
        def rename_key(key):
            if not 'module' in key:
                return key
            return ''.join(key.split('.module'))

        checkpoint['state_dict'] = {rename_key(key): val
                                    for key, val
                                    in checkpoint['state_dict'].items()}

        # load weights
        model.load_state_dict(checkpoint['state_dict'])
        print("Loaded")
    else:
        model = None
        print("=> no checkpoint found at '{}'".format(path))
    return model


class ExpressionDataset(Dataset):
    '''
    custom Dataset class for classifier training.
    __len__: returns the number of samples in our dataset.
    __getitem__:  loads and returns a sample from the dataset at the given index idx.
    '''
    def __init__(self, data_path):
        '''
        label_file: label
        '''
        self.datas = pd.read_csv(data_path, sep='\t', index_col=0)    ##load datas
        self.datas = np.expand_dims(np.array(self.datas), axis=1)
        self.labels = np.zeros((len(self.datas)), dtype='float32')
    
    def __len__(self):
        return len(self.datas)

    def __getitem__(self, idx):
        data = self.datas[idx]
        label = self.labels[idx]
        data = torch.FloatTensor(data)
        label = torch.LongTensor(np.array(label))
        return data, label
    
    
class ReassignedDataset(Dataset):
    """
    Dataset with reassigned labels based on clustering results.

    Parameters:
        data_indexes (list): List of data indexes.
        pseudolabels (list): List of pseudo labels.
        dataset (np.array): Original dataset.
    """
    def __init__(self, data_indexes, pseudolabels, dataset):
        self.datas, self.labels = self.make_dataset(data_indexes, pseudolabels, dataset)
    
    def make_dataset(self, data_indexes, pseudolabels, dataset):
        label_to_idx = {label: idx for idx, label in enumerate(set(pseudolabels))} 
        datas = np.zeros(dataset.shape, dtype='float32')
        labels = np.zeros((len(pseudolabels)), dtype='float32')
        for j, idx in enumerate(data_indexes):  
            data = dataset[idx]   
            pseudolabel = label_to_idx[pseudolabels[j]]
            datas[idx] = data    
            labels[idx] = pseudolabel
        return datas, labels     
    
    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        data = self.datas[idx]
        label = self.labels[idx]
        data = torch.FloatTensor(data)
        label = torch.LongTensor(np.array(label))
        return data, label
    
    
def pseudolabels_assign(datas_lists, dataset):
    """
    Creates a dataset from clustering, with clusters as labels.

    Args:
        data_lists (list of list): for each cluster, the list of data indexes
                                    belonging to this cluster
        dataset (list): initial dataset
    Returns:
        ReassignedDataset(torch.utils.data.Dataset): a dataset with clusters as
                                                     labels
    """
    assert datas_lists is not None
    pseudolabels = []
    data_indexes = []
    for cluster, datas in enumerate(datas_lists):
        data_indexes.extend(datas)
        pseudolabels.extend([cluster] * len(datas))
    return ReassignedDataset(data_indexes, pseudolabels, dataset)


class UnifLabelSampler(Sampler):
    """
    Samples elements uniformely accross pseudolabels.

    Args:
        N (int): size of returned iterator.
        images_lists: dict of key (target), value (list of data with this target)
    """

    def __init__(self, N, data_lists):
        self.N = N
        self.data_lists = data_lists
        self.indexes = self.generate_indexes_epoch()

    def generate_indexes_epoch(self):
        nmb_non_empty_clusters = 0
        for i in range(len(self.data_lists)):
            if len(self.data_lists[i]) != 0:
                nmb_non_empty_clusters += 1

        size_per_pseudolabel = int(self.N / nmb_non_empty_clusters) + 1
        res = np.array([])

        for i in range(len(self.data_lists)):
            # skip empty clusters
            if len(self.data_lists[i]) == 0:
                continue
            indexes = np.random.choice(
                self.data_lists[i],
                size_per_pseudolabel,
                replace=(len(self.data_lists[i]) <= size_per_pseudolabel)
            )
            res = np.concatenate((res, indexes))

        np.random.shuffle(res)
        res = list(res.astype('int'))
        if len(res) >= self.N:
            return res[:self.N]
        res += res[: (self.N - len(res))]
        return res

    def __iter__(self):
        return iter(self.indexes)

    def __len__(self):
        return len(self.indexes)
    
    
    
class AverageMeter(object):
    """
    Computes and stores the average and current value
    
    Methods:
        update(val, n): Update average with new value `val` and count `n`.    
    """
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count
        
    


import os
import sys
import time
import argparse
from tkinter import Variable
import numpy as np
import pandas as pd
import umap
import matplotlib.pyplot as plt
import torch.nn.functional as F
from torch.autograd import Variable
from sklearn.cluster import KMeans
from torch.optim import lr_scheduler

import torch
from torch import nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
from sklearn.metrics.cluster import normalized_mutual_info_score,homogeneity_score

from torch.utils.data import dataset
from sklearn.metrics import silhouette_score,calinski_harabasz_score,adjusted_rand_score

import models
import cluster
from utils import ExpressionDataset, pseudolabels_assign, UnifLabelSampler, AverageMeter

def parse_args():
    parser = argparse.ArgumentParser()
    
    # working_dir = os.path.dirname(os.path.abspath('main_copy2.py')) 
    parser.add_argument('--nmb_cluster', type=int, default=10,
                        help='number of cluster for k-means (default: 100)')
    parser.add_argument('--lr', type=float, default=0.01,
                        help='learning rate (default: 0.05)')
    parser.add_argument('--momentum', type=float, default=0.9,
                        help='momentum (default: 0.9)')
    parser.add_argument('--wd', default=-5, type=float,
                        help='weight decay pow (default: -5)')
    parser.add_argument('--data_path', metavar='PATH', help='path to dataset',
                        default='GeneCluster/data/normal_data.csv')

    parser.add_argument('--batch', type=int, default=128,
                        help='batch size')
    parser.add_argument('--workers', type=int, default=8,
                        help='number of data loading workers')
    parser.add_argument('--epochs', type=int, default=100,
                        help='epoch number')
    parser.add_argument('--clustering', type=str, choices=['Kmeans'],
                        default='Kmeans', help='clustering algorithm (default: Kmeans)')
    parser.add_argument('--arch', type=str, choices=['sppCNN'],
                        default='sppCNN', help='feature embedding model architecture')
    parser.add_argument('--reassign', type=float, default=1.,
                        help="""how many epochs of training between two consecutive
                        reassignments of clusters (default: 1)""")
    parser.add_argument('--ckpt_path', type=str, default='GeneCluster/train_res', help='')
    parser.add_argument('--checkpoints', type=int, default=1000,
                        help='how many iterations between two checkpoints (default: 1000)')
    parser.add_argument('--verbose', '-v', action='store_true',
                        default=True, help='verbose mode')
    
    return parser.parse_args()


def compute_features(dataloader, model, N):
    end = time.time()
    model.eval()  # close train model
    for batch, (input_tensor, _) in enumerate(dataloader):  
        with torch.no_grad():
            aux = model(input_tensor).data.cpu().numpy() 
        
        if batch == 0:
            features = np.zeros((N, aux.shape[1]), dtype='float32')  

        aux = aux.astype('float32')  
        if batch < len(dataloader) - 1:  
            features[batch * args.batch: (batch + 1) * args.batch] = aux
        else:

            features[batch * args.batch:] = aux

        time_cost = time.time() - end
        end = time.time()

        if (batch % 200) == 0:
            print('{0} / {1}\t'
                  'Time cost: {time:.3f} s'
                  .format(batch, len(dataloader), time=time_cost))
    return features     
    

def Visualization(features, data_lists=None): 
    """ Cluster visualization """
   
    reducer = umap.UMAP()
    embedding = reducer.fit_transform(features)

    if data_lists is None:
        kmeans = KMeans(n_clusters=args.nmb_cluster,random_state=10)     
        kmeans.fit(embedding)
        gene_label = kmeans.predict(embedding)
        
    else:
        gene_label = np.zeros(len(features))      
        for i in range(len(data_lists)):
            gene_label[data_lists[i]] = [i for m in range(len(data_lists[i]))]   

    
    picture = plt.scatter(embedding[:, 0], embedding[:, 1], s=5, c=gene_label)             
    plt.gca().set_aspect('equal', 'datalim')
    plt.colorbar(boundaries=np.arange(11)-0.5).set_ticks(np.arange(10))
    plt.title('UMAP projection of the dataset')
    # plt.gca().legend().remove()
    return picture


class FocalLoss(nn.Module):

    def __init__(self, alpha=None, gamma=2, reduction='mean'):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction
        
    def forward(self, output, targets):
        ce_loss = F.cross_entropy(output, targets, reduction='none')
        pt = torch.exp(-ce_loss)
        if self.alpha is not None:
            alpha_t = self.alpha[targets]
            focal_loss = alpha_t * (1 - pt) ** self.gamma * ce_loss
        else:
            focal_loss = (1 - pt) ** self.gamma * ce_loss
        if self.reduction == 'mean':
            return torch.mean(focal_loss)
        elif self.reduction == 'sum':
            return torch.sum(focal_loss)
        else:
            return focal_loss
        

def evaluat(vector,pred_label,true_label):
    # silhouette_avg = silhouette_score(vector,pred_label)
    # ch_score = calinski_harabasz_score(vector,pred_label)

    ari = adjusted_rand_score(true_label, pred_label)
    homo = homogeneity_score(true_label, pred_label)
    nmi = normalized_mutual_info_score(true_label, pred_label)
    return ari,homo,nmi


def main(args):
    # model
    # os.environ["CUDA_VISIBLE_DEVICES"] = "0"
    model = models.__dict__[args.arch](out=args.nmb_cluster)
    fea_dim = int(model.top_layer.weight.size()[1])  # torch.Size([100, 200])
    model.top_layer = None
    model.features = torch.nn.DataParallel(model.features)
    # device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    # model.features.to(device)
    model.cuda()
    cudnn.benchmark = True
    
    # load data
    end = time.time()
    dataset = ExpressionDataset(args.data_path)

    if args.verbose:
        print('Load dataset: {0:.2f} s'.format(time.time() - end))

    dataloader = torch.utils.data.DataLoader(dataset,
                                            batch_size=args.batch,
                                            num_workers=args.workers,
                                            pin_memory=True)


    deepcluster = cluster.__dict__[args.clustering](args.nmb_cluster) 
    if args.verbose:
            print('Cluster the features with {}'.format(args.clustering))
    
    
    # create optimizer
    optimizer = torch.optim.SGD(
        filter(lambda x: x.requires_grad, model.parameters()),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=10**args.wd,
    )

    scheduler = lr_scheduler.ExponentialLR(optimizer, gamma=0.9)

    # define loss function
    # loss_fn = nn.CrossEntropyLoss().cuda()
    # loss_fn = FocalLoss(args.nmb_cluster).cuda()
    loss_fn = FocalLoss(gamma=2, reduction='mean').cuda()
    
    # creating checkpoint repo
    ckpt_path = os.path.join(args.ckpt_path, 'checkpoints')
    if not os.path.isdir(ckpt_path):
        os.makedirs(ckpt_path)
    
    cluster_log = cluster.Logger(os.path.join(args.ckpt_path,'clusters'))

    # last_update = 0
    # patience = 20  
    for epoch in range(args.epochs):
        end = time.time()
        
        model.top_layer = None
        model.classifier = nn.Sequential(*list(model.classifier.children())[:-1]) 
        features = compute_features(dataloader, model, len(dataset))
        print(features.shape)
        # print(np.isnan(features))
        np.save(args.ckpt_path+'/features.npy',features)

        clustering_loss = deepcluster.run(features, verbose=args.verbose) 
        
        if epoch % 20 == 0:
            visual = Visualization(features, deepcluster.data_lists)

            plt.savefig(os.path.join(args.ckpt_path, 'Epoch{}_cluster_picture.png'.format(epoch)))
            plt.close()


        if args.verbose:
            print('Assign pseudo labels')            
        train_dataset = pseudolabels_assign(deepcluster.data_lists,
                                            dataset.datas)

        sampler = UnifLabelSampler(int(args.reassign * len(train_dataset)),   
                                   deepcluster.data_lists)
        
        train_dataloader = torch.utils.data.DataLoader(
            train_dataset,
            batch_size=args.batch,
            num_workers=args.workers,
            sampler=sampler,
            pin_memory=True)
        
        # set last fully connected layer
        mlp = list(model.classifier.children())
        mlp.append(nn.ReLU(inplace=True).cuda())  
        model.classifier = nn.Sequential(*mlp)
        model.top_layer = nn.Linear(fea_dim, len(deepcluster.data_lists))
        model.top_layer.weight.data.normal_(0, 0.01)
        model.top_layer.bias.data.zero_()
        model.top_layer.cuda()

         
        loss = train(train_dataloader, model, loss_fn, optimizer, epoch)
        scheduler.step()
        
        gene_list = pd.DataFrame({'clus_id':train_dataset.labels})
        gene_list['clus_id'] = gene_list['clus_id'].astype(int)
        gene_list.index = pd.read_csv(args.data_path,sep = '\t',index_col = 0).index
        gene_list.to_csv(args.ckpt_path+'/genelist.txt')

        # reducer = umap.UMAP()
        # vector = reducer.fit_transform(np.squeeze(train_dataset.datas))
        print(train_dataset.datas.shape)
        print(train_dataset.labels.shape)

        # pred_label = gene_list.loc[ref.gene].clus_id
        # true_label = ref.group
        # ari,homo,nmi = evaluat(train_dataset.datas,pred_label,true_label)
        # print('ari: {:.3f}\t' 'homo: {:.3f}\t' 'nmi: {:.3f}\n'
        #        .format(ari,homo,nmi))
        
        silhouette_avg = silhouette_score(features,train_dataset.labels)
        ch_score = calinski_harabasz_score(features,train_dataset.labels)
        print('silhouette_avg: {:.3f}\n'
              'ch_score: {:.3f}\n'
              .format(silhouette_avg,ch_score))


        # print log
        if args.verbose:
            print('###### Epoch [{0}] ###### \n'
                  'Time: {1:.3f} s\n'
                  'Clustering loss: {2:.3f} \n'
                  'ConvNet loss: {3:.3f}\n'
                  .format(epoch, time.time() - end, clustering_loss, loss))
            
            try:
                NMI = normalized_mutual_info_score(
                        cluster.arrange_clustering(deepcluster.data_lists),
                        cluster.arrange_clustering(cluster_log.data[-1])
                    )
                print('NMI against previous assignment: {0:.3f}'.format(NMI))
            except IndexError:
                pass

            cluster_log.log(deepcluster.data_lists)
            #print(cluster_log.data)
            print('####################### \n')
            
        # save running checkpoint
        torch.save({'epoch': epoch + 1,
		            'arch': args.arch,
                    'clustering': args.clustering,
                    'state_dict': model.state_dict(),
                    'optimizer' : optimizer.state_dict()},
                   os.path.join(args.ckpt_path, 'checkpoint.pth.tar'))

        # # early stopping
        # if epoch == 0:
        #     best_loss = loss
        # else:
        #     if loss < best_loss:
        #         best_loss = loss
        #         last_update = epoch
        #     else:
        #         if epoch - last_update > patience:
        #             print("Early stopping!")
        #             break        


    
def train(dataloader, model, loss_fn, optimizer, epoch):

    losses = AverageMeter()
    batch_time = AverageMeter()
    
    # switch to train mode
    model.train()

    # create an optimizer for the last fc layer
    optimizer_tl = torch.optim.SGD(
        model.top_layer.parameters(),
        lr=args.lr,
        weight_decay=10**args.wd,
    )

    for batch, (input_tensor, target) in enumerate(dataloader):
        end = time.time()

        # save checkpoint
        n = len(dataloader) * epoch + batch
        if n % args.checkpoints == 0:
            path = os.path.join(
                args.ckpt_path,
                'checkpoints',
                'checkpoint_' + str(n / args.checkpoints) + '.pth.tar',
            )
            if args.verbose:
                print('Save checkpoint at: {0}'.format(path))
            torch.save({
                'epoch': epoch + 1,
		        'arch': args.arch,
                'clustering': args.clustering,
                'state_dict': model.state_dict(),
                'optimizer' : optimizer.state_dict()
            }, path)

        target = target.cuda()
        input_var = torch.autograd.Variable(input_tensor.cuda())
        target_var = torch.autograd.Variable(target)

        output = model(input_var)
        # probs = F.softmax(output,dim=1)
        # # pred = np.argmax(probs)
        # print(probs)

        # loss = loss_fn(output, target_var)
        loss = loss_fn(output, target_var)

        # record loss
        losses.update(loss.item(), input_tensor.size(0))

        # compute gradient and do SGD step
        optimizer.zero_grad()
        optimizer_tl.zero_grad()
        loss.backward()    
        optimizer.step()
        optimizer_tl.step()

        # measure elapsed time
        batch_time.update(time.time() - end)

        if args.verbose and (batch % 200) == 0:
            print('Epoch: [{0}][{1}/{2}]\t'
                  'Time: {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                  'Loss: {loss.val:.4f} ({loss.avg:.4f})'
                  .format(epoch, batch, len(dataloader),
                          batch_time=batch_time,
                          loss=losses))

    return losses.avg
    
if __name__ == '__main__':
    end1 = time.time()
    args = parse_args()
    sys.stdout = open(args.ckpt_path + '/result.log', mode = 'w')
    main(args)
    print((time.time() - end1)/60)



from setuptools import setup, find_packages

setup(
    name="GeneCluster",
    version="0.1.1",
    author="Byting",
    author_email="yutingya820@163.com",
    description="A gene clustering model using deep learning",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Byting820/GeneCluster",
    packages=find_packages(),
    install_requires=[
        "torch",
        "torchvision",
        "numpy",
        "pandas",
        "umap-learn",
        "matplotlib",
        "scikit-learn",
        "argparse"
    ],
    entry_points={
        "console_scripts": [
            "GeneCluster=genecluster.main:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)


import os
import pickle
import time
import faiss
import numpy as np

import torch
from torch.utils.data import Dataset


__all__ = ['Kmeans']

def preprocess_features(npdata, pca=20):
    """ Applies PCA for dimensionality reduction, whitening, and L2 normalization.. 
    Preprocess an array of features.

    Args:
        npdata (np.array N * ndim): features to preprocess
        pca (int): dim of output
    Returns:
        np.array of dim N * pca: data PCA-reduced, whitened and L2-normalized
    """
    _, ndim = npdata.shape
    npdata =  npdata.astype('float32')

    # Apply PCA-whitening with Faiss
    mat = faiss.PCAMatrix(ndim, pca, eigen_power=-0.5) 
    mat.train(npdata) 
    assert mat.is_trained
    npdata = mat.apply_py(npdata)

    # L2 normalization
    row_sums = np.linalg.norm(npdata, axis=1) 
    npdata = npdata / row_sums[:, np.newaxis]

    return npdata   


def run_kmeans(x, nmb_clusters, verbose=False):
    """Runs kmeans on GPU.

    Args:
        x: data
        nmb_clusters (int): number of clusters
    Returns:
        list: ids of data in each cluster
    """
    n_data, d = x.shape

    # faiss implementation of k-means
    clus = faiss.Clustering(d, nmb_clusters)

    # Change faiss seed at each k-means so that the randomly picked
    # initialization centroids do not correspond to the same feature ids
    # from an epoch to another.
    # clus.seed = np.random.randint(1234) 

    clus.niter = 20 

    clus.max_points_per_centroid = 1000   
     
    res = faiss.StandardGpuResources()
    flat_config = faiss.GpuIndexFlatConfig()
    flat_config.useFloat16 = False
    flat_config.device = 0
    index = faiss.GpuIndexFlatL2(res, d, flat_config)

    # perform the training
    clus.train(x, index) 
    _, I = index.search(x, 1)  
    stats = clus.iteration_stats
    losses = np.array([stats.at(i).obj for i in range(stats.size())])
    if verbose:
        print('k-means loss evolution: {0}'.format(losses))

    return [int(n[0]) for n in I], losses[-1]


class Kmeans(object):
    """
    K-means clustering class for applying k-means on data with preprocessing.

    Attributes:
        nmb_clusters (int): Number of clusters.
    """
    def __init__(self, nmb_clusters):
        self.nmb_clusters = nmb_clusters

    def run(self, data, verbose=False):
        """Performs k-means clustering.
            Args:
                x_data (np.array N * dim): data to cluster
        """
        end = time.time()
        
        # PCA-reducing, whitening and L2-normalization
        self.xb = preprocess_features(data)

        # cluster the data
        I, loss = run_kmeans(self.xb, self.nmb_clusters, verbose) 
        self.I = I
        
        self.data_lists = [[] for i in range(self.nmb_clusters)]
        for i in range(len(data)):
            self.data_lists[I[i]].append(i)

        if verbose:
            print('k-means time: {0:.0f} s'.format(time.time() - end))

        return loss
    
    
class Logger(object):
    """ 
    Class to update every epoch to keep trace of the results.

    Methods:
        - log() log and save
    """

    def __init__(self, path):
        self.path = path
        self.data = []

    def log(self, train_point):
        """
        Logs a new training point and saves it to a file.

        Args:
            train_point (dict): Training data point to log.
        """
        self.data.append(train_point)
        with open(os.path.join(self.path), 'wb') as fp:
            pickle.dump(self.data, fp, -1)   


def arrange_clustering(data_lists):
    """
    Arranges clustering results by assigning pseudolabels based on cluster indices.

    Args:
        data_lists (List[List[int]]): List of clusters, where each sublist contains indices of data points in that cluster.

    Returns:
        np.ndarray: Array of pseudolabels aligned with original data order.
    """
    pseudolabels = []
    data_indexes = []
    for cluster, datas in enumerate(data_lists):
        data_indexes.extend(datas)
        pseudolabels.extend([cluster] * len(datas))
    indexes = np.argsort(data_indexes)  
    return np.asarray(pseudolabels)[indexes]    
    
